package metrics

import "time"

// Noop statter-- useful for unit testing.
type Noop struct{}

func (*Noop) Increment(metricName string, dimensions ...string)                      {}
func (*Noop) IncrementAt(metricName string, ts time.Time, dimensions ...string)      {}
func (*Noop) IncrementErr(metricName string, errDimension string)                    {}
func (*Noop) IncrementErrAt(metricName string, errDimension string, ts time.Time)    {}
func (*Noop) Timing(metricName string, duration time.Duration, dimensions ...string) {}
func (*Noop) TimingAt(metricName string, duration time.Duration, ts time.Time, dimensions ...string) {
}
func (*Noop) StartPoller(metricName string, valueFn func() float64, interval time.Duration) {}
func (*Noop) Size(metricName string, size int, dimensions ...string)                        {}
func (*Noop) Close() error                                                                  { return nil }
